from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import StrEnum


class Currency(StrEnum):
    EUR = "EUR"
    USD = "USD"


@dataclass
class Concept:
    """This class represent a potential match or the raw concept.

    The parser may try to generate the `best_concept_match` from the raw concept.
    If the `best_concept_match` is not useful for the categorizer, it can choose
    to use the raw concept instead.
    """

    best_concept_match: str
    raw: str


@dataclass
class Tx:
    """Represents a transaction extracted from a bank"""

    date: datetime
    amount: Decimal
    currency: Currency
    concept: str | Concept
    category: None | str
    metadata: dict[str, str]
    tags: list[str]



import features


def preprocess(inp: dict):
    """
    Each preprocessing step takes a dict and returns a dict.
    It is assumed that each steps knows how to deal with the content
    of the input received.
    """
    steps = (
        features.make_lowercase,
        features.remove_numbers,
        features.has_iban,
        features.extract_city,
        features.extract_payment_method,
    )
    out = inp
    for func in steps:
        out = func(out)
    return out


preproc = preprocess({"desc": "HELLO world"})
feats = features.get_features(preproc)
encoded = features.one_hot_encode(feats)



"""ABN AMRO is a Dutch Bank

Even though, ABN operates in more countries, we are only covering the NL branch with xls files.
"""

import re
from datetime import datetime
from decimal import Decimal
from pathlib import Path

import xlrd  # type: ignore

from ficamp.datastructures import Currency, Tx
from ficamp.parsers.protocols import Parser

TRANSACTIONDATE_REGEX = r"(\d{4})(\d{2})(\d{2})"
transactiondate_re = re.compile(TRANSACTIONDATE_REGEX)


def transactiondate_parser(value: str) -> datetime:
    """Parses an ABN string date into a datetime."""
    out = transactiondate_re.match(value)
    if out is None:
        raise ValueError("No proper datetime format")
    year, month, day = out.groups()
    return datetime(int(year), int(month), int(day))


def amount_parser(value: str) -> Decimal:
    """Parses a string into a Decimal."""
    v = value.replace(",", ".")
    return Decimal(v)


class ConceptParser:
    """Parse ABN concepts.

    Concept samples:

    ```
    BEA, Betaalpas                   Hortus Botanicus,PAS172         NR:1W5P01, 30.12.23/17:48        AMSTERDAM
    BEA, Betaalpas                   Gamma Compact,PAS172            NR:7V0GM0, 15.01.24/18:15        AMSTERDAM
    SEPA Overboeking                 IBAN: XXXXX        BIC: ABNANL2A                    Naam: ABN AMRO INZAKE GELDMAAT  Omschrijving: Storting 12-01-24  15:36 uur, Pas 172              Geldautomaat 812936              1053EK Amsterdam                Kenmerk: 401215363553
    ```

    """

    TRTP_REGEX = r"/(NAME|REMI|CSID)/([^/]+)"
    trtp_re = re.compile(TRTP_REGEX)

    SEPA_REGEX = r"Naam:\s(.*)\s{2,}Omschrijving:\s(.*)"
    sepa_re = re.compile(SEPA_REGEX)

    BEA_REGEX = r"(\s{2,})(.+?),PAS\d+\s+\S+\s+\S+\s+(\w+)"
    bea_re = re.compile(BEA_REGEX)

    def parse(self, concept: str) -> str:
        parsers = {
            "DEPOSIT INV": self.parse_deposit_inv,
            "BEA": self.parse_bea_gea,
            "GEA": self.parse_bea_gea,
            "/TRTP": self.parse_trtp,
            "ABN": self.parse_abn,
            "SEPA": self.parse_sepa,
            "KOOP": self.parse_abn,
        }
        for key, parser in parsers.items():
            if concept.startswith(key):
                # print(f"Using {parser}")
                try:
                    return parser(concept)
                except ValueError:
                    continue

        return "UNKNOWN"

    def parse_trtp(self, concept: str) -> str:
        """TRTP is a very ugly ugly thing.

        It's a string with a lot of fields separated by slashes.
        """
        matches = self.trtp_re.findall(concept)
        if not matches:
            raise ValueError("[TRTP] No matches found")
        out = "|".join([f"{key}:{value}" for key, value in matches])
        return out

    def parse_sepa(self, concept: str) -> str:
        matches = self.sepa_re.search(concept)
        if not matches:
            raise ValueError("[SEPA] No matches found")
        out = "|".join(["SEPA"] + list(matches.groups()))
        return out

    def parse_bea_gea(self, concept: str) -> str:
        match = self.bea_re.search(concept)
        if not match:
            raise ValueError("[BEA/GEA] No matches found")
        return f"{match.group(2)}|{match.group(3)}"

    def parse_deposit_inv(self, concept: str) -> str:
        return "Investments ABN|" + concept

    def parse_abn(self, concept: str) -> str:
        return "ABN AMRO|" + concept


class AbnParser(Parser):
    """Parser for ABN AMRO bank statements.

    ABN uses the old `.xls` excel version, and we require xlrd
    to properly parse those kind of files
    """

    def __init__(self):
        self.concept_parser = ConceptParser()

    def load(self, filename: Path):
        self.book = xlrd.open_workbook(filename)

    def parse(self) -> list[Tx]:
        """ABN XLS parser.

        Columns by index in file:
            0. accountNumber
            1. mutationcode -> currency
            2. transactiondate -> date
            3. valuedate
            4. startsaldo
            5. endsaldo
            6. amount -> Decimal
            7. description -> str
        """
        book = self.book
        sheet = book.sheet_by_index(0)

        txs = []
        for i in range(1, sheet.nrows):
            currency = sheet.cell_value(i, 1)
            date = str(sheet.cell_value(i, 2))
            amount = str(sheet.cell_value(i, 6))
            concept = sheet.cell_value(i, 7)
            print("currency", currency)
            print("date", date)
            print("amount", amount)
            print("concept", concept)
            tx = self.build_transaction(date, amount, currency, concept)
            txs.append(tx)
        return txs

    def build_transaction(
        self,
        date: str,
        amount: str,
        currency: str,
        concept: str,
    ) -> Tx:
        _date = transactiondate_parser(date)
        _amount = amount_parser(amount)
        _currency = Currency(currency)

        _concept = self.concept_parser.parse(concept)

        return Tx(
            date=_date,
            amount=_amount,
            currency=_currency,
            concept=_concept,
            category=None,
            metadata={},
            tags=[],
        )



from datetime import datetime
from decimal import Decimal
from pathlib import Path

from openpyxl import load_workbook

from ficamp.datastructures import Tx
from ficamp.parsers.protocols import Parser


class AccountBBVAParser(Parser):
    """Parser for BBVA bank account extract"""

    def load(self, filename: Path | None = None):
        # TODO: rearrange this.

        # filename = Path("../data/enero-febrero-bbva-cuenta.xlsx")
        # filename = os.path.join(
        #     os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        #     "data/enero-febrero-bbva-cuenta.xlsx",
        # )

        wb = load_workbook(filename)
        sheet = wb.active
        start_row = 6
        start_column = 2

        return [
            row
            for row in sheet.iter_rows(
                min_row=start_row, min_col=start_column, values_only=True
            )
        ]

    def parse(self) -> list[Tx]:
        rows = self.load()

        return [
            self.row_processor(row)
            for row in rows
            if self.row_processor(row) is not None
        ]

    def row_processor(self, row):
        # Skip Credit Card charge in Account
        if "targeta" in row[2] or "tarjeta" in row[2]:
            return None

        if row[5] == "EUR":
            currency = "€"

        concept = f"{row[2]} || {row[3]}"

        return Tx(
            date=row[0],
            amount=Decimal(str(row[4])),
            currency=currency,
            concept=concept,
            category=None,
            metadata={"more_details": row[8], "origin": "BBVA Account"},
            tags=[],
        )


class CreditCardBBVAParser(Parser):
    """Parser for BBVA Credit Card Extract"""

    def load(self, filename: Path | None = None):
        # TODO: rearrange this
        # filename = Path("../data/enero-febrero-bbva-cuenta.xlsx")
        # filename = os.path.join(
        #     os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        #     "data/enero-febrero-bbva-targeta.xlsx",
        # )

        wb = load_workbook(filename)
        sheet = wb.active
        start_row = 6
        start_column = 2

        return [
            row
            for row in sheet.iter_rows(
                min_row=start_row, min_col=start_column, values_only=True
            )
        ]

    def parse(self) -> list[Tx]:
        rows = self.load()

        return [
            self.row_processor(row)
            for row in rows
            if self.row_processor(row) is not None
        ]

    def row_processor(self, row):
        # Skip Recharging the Credit Card, as it's an useless operation from user's POV.
        if row[3] > 0:
            return None

        currency = "€"

        return Tx(
            date=datetime.strptime(row[0], "%d/%m/%Y"),
            amount=Decimal(str(row[3])),
            currency=currency,
            concept=row[2],
            category=None,
            metadata={"origin": "BBVA Credit Card"},
            tags=[],
        )


if __name__ == "__main__":
    bbva = AccountBBVAParser()
    txs = bbva.parse()

    txs += CreditCardBBVAParser().parse()

    for tx in txs:
        print(tx)



import json
import os

import requests

GOOGLE_API_KEY = os.getenv("GOOGLE_MAPS_API_KEY", "FIXME_CONFIGURE_API_KEY")


class GoogleException(Exception):
    """Custom Exception for raising errors to the caller."""


def search_google_maps(business_name, location=None, api_key=GOOGLE_API_KEY):
    base_url = "https://maps.googleapis.com/maps/api/place/textsearch/json"
    params = {"query": business_name, "key": api_key}
    if location:
        params["location"] = location
        params["radius"] = 5000  # Radius in meters (you can adjust this)

    response = requests.get(base_url, params=params)
    response.raise_for_status()
    if response.json()["status"] != "OK":
        raise GoogleException(response.json()["error_message"])
    results = response.json().get("results", [])
    if results:
        # Assuming the first result is the most relevant
        categories = results[0].get("types", [])
        place_id = results[0].get("place_id", None)
        return place_id, categories
    return None, None


def get_place_details(place_id):
    url = f"https://places.googleapis.com/v1/places/{place_id}"
    headers = {
        "Content-Type": "application/json",
        "X-Goog-Api-Key": GOOGLE_API_KEY,
        "X-Goog-FieldMask": "id,displayName,types",
    }
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    if response.json()["status"] != "OK":
        raise GoogleException(response.json()["error_message"])
    return response.json().get("types", [])


def query_google_places_new(query):
    url = "https://places.googleapis.com/v1/places:searchText"
    headers = {
        "Content-Type": "application/json",
        "X-Goog-Api-Key": GOOGLE_API_KEY,
        "X-Goog-FieldMask": "places.name,places.types,places.formattedAddress",
    }
    payload = {"textQuery": query}
    response = requests.post(url, headers=headers, data=json.dumps(payload))
    response.raise_for_status()
    if response.json()["status"] != "OK":
        raise GoogleException(response.json()["error_message"])
    places = response.json().get("places", [])
    if places:
        categories = places[0].get("types", [])
        place_id = places[0].get("place_id", None)
        return place_id, categories
    return None, None


def find_business_category_in_google(field, location=None):
    keys_to_remove = ["point_of_interest", "establishment", "store", "department_store"]
    # first try using google map places search
    place_id_gmaps, categories = search_google_maps(field, location)
    if not categories and place_id_gmaps:
        # if we get a place_id but no categories then query the details
        categories = get_place_details(place_id_gmaps)
    if not categories:
        # if still no categories then try with the new API
        place_id_gplaces, categories = query_google_places_new(field)
        if not categories and place_id_gplaces:
            # if we get a place_id but no categories then query the details
            categories = get_place_details(place_id_gplaces)
    if not categories:
        raise GoogleException("Not found in G Maps or G Places")
    categories = list(set(categories) - set(keys_to_remove))
    if categories:
        categories = list(set(categories) - set(keys_to_remove))
        return categories[0]
    raise GoogleException



import argparse

from dotenv import load_dotenv

from ficamp.parsers.abn import AbnParser


def cli() -> argparse.Namespace:
    """Parses the first argument from the command line and prints it."""

    # Create an argument parser
    parser = argparse.ArgumentParser(
        prog="ficamp", description="Print the first argument from the CLI"
    )

    parser.add_argument("--bank", choices=["abn"], default="abn")
    parser.add_argument("filename", help="The spreadsheet to load")

    # Parse the arguments
    args = parser.parse_args()

    # Print the first argument
    return args


def main():
    args = cli()
    args.filename
    args.bank

    # TODO: Build enum for banks
    if args.bank == "abn":
        parser = AbnParser()
        parser.load(args.filename)
        transactions = parser.parse()
        print(transactions)
        # TODO: Add categorizer!


load_dotenv()
main()



from pathlib import Path
from typing import Protocol

from ficamp.datastructures import Tx


class Parser(Protocol):
    """Main protocol to define parsers for different banks."""

    def load(self, filename: Path): ...

    def parse(self) -> list[Tx]: ...



from dataclasses import dataclass
from functools import lru_cache
from typing import Any

import numpy as np
from sklearn.utils import murmurhash3_32


@dataclass
class HashFeature:
    cardinality: int


def sort_dict(d):
    return dict(sorted(d.items(), key=lambda x: x[0]))


@lru_cache(500)
def categorical_hash_bucket(item: str, *, buckets: int) -> int:
    """Integer representation of item achieved like this: hash(item) % buckets"""
    if not isinstance(item, str):
        raise ValueError("item must be a string")
    return murmurhash3_32(item, positive=True) % buckets


def make_lowercase(d):
    return {"desc": d["desc"].lower()}


def remove_numbers(d):
    """Return string without number like SEPA 12312321 BIC --> SEPA BIC"""
    # TODO: remove hardcoded value
    return d


def has_iban(d) -> str:
    """
    Return str(int(True)) if IBAN detected, else str(int(False)).
    str because we need string to encode feature"""
    # TODO: remove hardcoded value
    return d | {"has_iban": str(int(True))}


def extract_city(d) -> str:
    "Return city name in lowercase if found, else <UNK>"
    # TODO: remove hardcoded value
    return d | {"city": "Lleida"}


def extract_payment_method(d: dict) -> str | dict[str, Any]:
    "Return payment method name in lowercase if found, else <UNK>"
    # TODO:improve logic

    payment_methods = sorted(
        (
            "card",
            "creditcard",
            "paypal",
            "transfer",
        )
    )
    res = "<UNK>"
    for method in payment_methods:
        if method in d["desc"]:
            return method
    return d | {"payment_method": res}


# This config determines which features will be extracted
# from the transaction description ("desc")
CONFIG = {
    "city": HashFeature(cardinality=100),
    "has_iban": HashFeature(cardinality=2),
    "payment_method": HashFeature(cardinality=10),
}


def get_features(preprocessed: dict):
    features = {}
    for name, conf in CONFIG.items():
        features[name] = categorical_hash_bucket(
            preprocessed[name], buckets=conf.cardinality
        )
    return features


def one_hot_encode(features: dict):
    encoded = {}
    for name, idx in sort_dict(features).items():
        c = CONFIG[name].cardinality
        one_hot = np.zeros(c)
        one_hot[idx] = 1
        encoded[name] = one_hot
    return encoded



import os
from datetime import datetime
from decimal import Decimal
from pathlib import Path

from openpyxl import load_workbook

from ficamp.datastructures import Tx
from ficamp.parsers.protocols import Parser


class AccountBSabadellParser(Parser):
    """Parser for BBVA bank account extract"""

    def load(self, filename: Path | None = None):
        # TODO: rearrange this.

        # filename = Path("../data/enero-febrero-bbva-cuenta.xlsx")
        # filename = os.path.join(
        #     os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        #     "data/enero-febrero-bsabadell-cuenta.xlsx",
        # )

        wb = load_workbook(filename)
        sheet = wb.active
        start_row = 10
        start_column = 1

        return [
            row
            for row in sheet.iter_rows(
                min_row=start_row, min_col=start_column, values_only=True
            )
        ]

    def parse(self) -> list[Tx]:
        rows = self.load()

        return [
            self.row_processor(row)
            for row in rows
            if self.row_processor(row) is not None
        ]

    def row_processor(self, row):
        # Skip Credit Card charge in Account
        if "TARJETA CREDITO" in row[2]:
            return None

        currency = "€"

        concept = self.concept_builder(row)

        return Tx(
            date=datetime.strptime(row[0], "%d/%m/%Y"),
            amount=Decimal(str(row[3])),
            currency=currency,
            concept=concept,
            category=None,
            metadata={"origin": "BSABADELL Account"},
            tags=[],
        )

    def concept_builder(self, row):
        """
        There are some rows with more details than others.
        This function builds the concept with all the details available.
        """
        concept = row[1]
        if row[5]:
            concept += f" || {row[5]}"
        if row[6]:
            concept += f" || {row[6]}"

        return concept


class CreditCardBSabadellParser(Parser):
    """Parser for Banc Sabadell Credit Card Extract"""

    def load(self, filename: Path | None = None):
        # TODO: rearrange this
        # filename = Path("../data/enero-febrero-bbva-cuenta.xlsx")
        # filename = os.path.join(
        #     os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        #     "data/enero-febrero-bsabadell-targeta.xlsx",
        # )

        wb = load_workbook(filename)
        sheet = wb.active
        start_row = 12
        start_column = 1

        rows = []
        for row in sheet.iter_rows(
            min_row=start_row, min_col=start_column, values_only=True
        ):
            if row[0] is None:
                # Mixed content for xlsx file, so we need to break when we reach the end of the table
                break

            rows.append(row)
        return rows

    def parse(self) -> list[Tx]:
        rows = self.load()
        return [
            self.row_processor(row)
            for row in rows
            if self.row_processor(row) is not None
        ]

    def row_processor(self, row):
        currency = "€"

        return Tx(
            date=datetime.strptime(f"{row[0]}/{datetime.now().year}", "%d/%m/%Y"),
            amount=Decimal(str(row[4]).replace(",", ".")),
            currency=currency,
            concept=row[1],
            category=None,
            metadata={"origin": "BSABADELL Credit Card", "location": row[2]},
            tags=[],
        )


if __name__ == "__main__":
    bsabadell = AccountBSabadellParser()
    txs = bsabadell.parse()

    txs += CreditCardBSabadellParser().parse()

    for tx in txs:
        print(tx.concept)
